@@ -1,45 +1,10 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 # Author: Xie Yanbo <xieyanbo@gmail.com>
-# Author: Louie Lu <louie.lu@hopebaytech.com>
 # This software is licensed under the New BSD License. See the LICENSE
 # file in the top distribution directory for the full license text.
 
-"""A debug library for RobotFramework, which can be used as an interactive
-shell(REPL) also.
-
-As a library:
-*** Settings ***
-Library         DebugLibrary
-
-** test case **
-SOME TEST
-    # some keywords...
-    Debug
-
-Run standalone:
-$ python DebugLibrary.py
-[...snap...]
->>>>> Enter interactive shell
-Only accepted plain text format keyword seperated with two or more spaces.
-Type "help" for more information.
-> log  hello
-> get time
-< '2016-07-20 11:51:33'
-> import library  String
-> get substring  helloworld  5  8
-< 'wor'
-> ${secs} =  Get Time  epoch
-# ${secs} = 1474814470
-> Log to console  ${secs}
-1474814470
-> selenium  google.com  chrome
-# import library  Selenium2Library
-# open browser  http://google.com  chrome
-< 1
-> close all browsers
-> Ctrl-D
->>>>> Exit shell.
+"""A debug library and REPL for RobotFramework.
 """
 
 from __future__ import print_function
@@ -48,6 +13,7 @@ import cmd
 import os
 import re
 import sys
+import tempfile
 from functools import wraps
 
 from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
@@ -58,6 +24,8 @@ from prompt_toolkit.shortcuts import print_tokens, prompt
 from prompt_toolkit.styles import style_from_dict
 from pygments.token import Token
 
+from robot import version
+from robot import run_cli
 from robot.api import logger
 from robot.errors import ExecutionFailed, HandlerExecutionFailed
 from robot.libdocpkg.model import LibraryDoc
@@ -65,9 +33,17 @@ from robot.libdocpkg.robotbuilder import KeywordDocBuilder, LibraryDocBuilder
 from robot.libraries import STDLIBS
 from robot.libraries.BuiltIn import BuiltIn
 from robot.running.namespace import IMPORTER
-from robot.variables import is_var
+from robot.running.signalhandler import STOP_SIGNAL_MONITOR
 
-__version__ = '1.0.3'
+rf_version = version.get_version(naked=False)
+rf_version = float(rf_version[:3])
+
+if rf_version <= 3.1:
+    from robot.variables import is_var    
+elif rf_version >= 3.2:
+    from robot.variables.search import is_variable
+
+__version__ = '1.3.0'
 
 HISTORY_PATH = os.environ.get('RFDEBUG_HISTORY', '~/.rfdebug_history')
 KEYWORD_SEP = re.compile('  +|\t')
@@ -134,6 +110,11 @@ def get_libs():
     return sorted(IMPORTER._library_cache._items, key=lambda _: _.name)
 
 
+def get_libs_as_dict():
+    """Get libraries robotframework imported as a name -> lib dict"""
+    return {l.name: l for l in IMPORTER._library_cache._items}
+
+
 def match_libs(name=''):
     """Find libraries by prefix of library name, default all"""
     libs = [_.name for _ in get_libs()]
@@ -146,12 +127,13 @@ def memoize(function):
     memo = {}
 
     @wraps(function)
-    def wrapper(*args):
-        if args in memo:
-            return memo[args]
+    def wrapper(*args, **kwargs):
+        key = (args, frozenset(sorted(kwargs.items())))
+        if key in memo:
+            return memo[key]
         else:
-            rv = function(*args)
-            memo[args] = rv
+            rv = function(*args, **kwargs)
+            memo[key] = rv
             return rv
     return wrapper
 
@@ -168,12 +150,15 @@ class ImportedLibraryDocBuilder(LibraryDocBuilder):
 
 
 @memoize
-def get_lib_keywords(library):
+def get_lib_keywords(library, long_format=False):
     """Get keywords of imported library"""
     lib = ImportedLibraryDocBuilder().build(library)
     keywords = []
     for keyword in lib.keywords:
-        doc = keyword.doc.split('\n')[0]
+        if long_format:
+            doc = keyword.doc
+        else:
+            doc = keyword.doc.split('\n')[0]
         keywords.append({'name': keyword.name,
                          'lib': library.name,
                          'doc': doc})
@@ -213,29 +198,69 @@ def print_error(head, message, style=ERROR_STYLE):
     print_output(head, message, style=style)
 
 
+def parse_keyword(command):
+    unicode_command = ''
+    if sys.version_info > (3,):
+        unicode_command = command
+    else:
+        unicode_command = command.decode(COMMAND_LINE_ENCODING)
+    return KEYWORD_SEP.split(unicode_command)
+
+
+def assign_variable(bi, variable_name, args):
+    variable_value = bi.run_keyword(*args)
+    bi._variables.__setitem__(variable_name, variable_value)
+    return variable_value
+
+
 def run_keyword(bi, command):
     """Run a keyword in robotframewrk environment"""
     if not command:
         return
     try:
-        u_command = ''
-        if sys.version_info > (3,):
-            u_command = command
-        else:
-            u_command = command.decode(COMMAND_LINE_ENCODING)
-        keyword = KEYWORD_SEP.split(u_command)
-        variable_name = keyword[0].rstrip('= ')
+        keyword_args = parse_keyword(command)
+        keyword = keyword_args[0]
+        args = keyword_args[1:]
 
-        if is_var(variable_name):
-            variable_value = bi.run_keyword(*keyword[1:])
-            bi._variables.__setitem__(variable_name,
-                                      variable_value)
-            print_output('#',
-                         '{} = {!r}'.format(variable_name, variable_value))
-        else:
-            result = bi.run_keyword(*keyword)
-            if result:
-                print_output('<', repr(result))
+        is_comment = keyword.strip().startswith('#')
+        if is_comment:
+            return
+
+        variable_name = keyword.rstrip('= ')
+        if rf_version <= 3.1:
+            if is_var(variable_name):
+            variable_only = not args
+            if variable_only:
+                display_value = ['Log to console', keyword]
+                bi.run_keyword(*display_value)
+            else:
+                variable_value = assign_variable(bi,
+                                                 variable_name,
+                                                 args)
+                print_output('#',
+                             '{} = {!r}'.format(variable_name,
+                                                variable_value))
+            else:
+                result = bi.run_keyword(keyword, *args)
+                if result:
+                    print_output('<', repr(result))
+        elif rf_version >= 3.2:
+            if is_variable(variable_name):
+            variable_only = not args
+            if variable_only:
+                display_value = ['Log to console', keyword]
+                bi.run_keyword(*display_value)
+            else:
+                variable_value = assign_variable(bi,
+                                                 variable_name,
+                                                 args)
+                print_output('#',
+                             '{} = {!r}'.format(variable_name,
+                                                variable_value))
+            else:
+                result = bi.run_keyword(keyword, *args)
+                if result:
+                    print_output('<', repr(result))
     except ExecutionFailed as exc:
         print_error('! keyword:', command)
         print_error('!', exc.message)
@@ -312,7 +337,7 @@ class PtkCmd(BaseCmd):
     get_prompt_tokens = None
     prompt_style = None
     intro = '''\
-Only accepted plain text format keyword seperated with two or more spaces.
+Only accepted plain text format keyword separated with two or more spaces.
 Type "help" for more information.\
 '''
 
@@ -344,6 +369,9 @@ Type "help" for more information.\
         cmd_completer = CmdCompleter(commands, self)
         return cmd_completer
 
+    def pre_loop(self):
+        pass
+
     def cmdloop(self, intro=None):
         """Better command loop supported by prompt_toolkit
 
@@ -356,6 +384,7 @@ Type "help" for more information.\
 
         stop = None
         while not stop:
+            self.pre_loop()
             if self.cmdqueue:
                 line = self.cmdqueue.pop(0)
             else:
@@ -403,6 +432,15 @@ class DebugCmd(PtkCmd):
         """Run after a command"""
         return stop
 
+    def reset_robotframework_exception(self):
+        if STOP_SIGNAL_MONITOR._signal_count:
+            STOP_SIGNAL_MONITOR._signal_count = 0
+            STOP_SIGNAL_MONITOR._running_keyword = True
+            logger.info('Reset last exception by DebugLibrary')
+
+    def pre_loop(self):
+        self.reset_robotframework_exception()
+
     def do_help(self, arg):
         """Show help message."""
         if not arg.strip():
@@ -446,14 +484,14 @@ use the TAB keyboard key to autocomplete keywords.\
         return cmd_completer
 
     def do_selenium(self, arg):
-        """Start a selenium 2 webdriver and open url in browser you expect.
+        """Start a selenium webdriver and open url in browser you expect.
 
         s(elenium)  [<url>]  [<browser>]
 
         default url is google.com, default browser is firefox.
         """
 
-        command = 'import library  Selenium2Library'
+        command = 'import library  SeleniumLibrary'
         print_output('#', command)
         run_keyword(self.rf_bi, command)
 
@@ -535,9 +573,11 @@ use the TAB keyboard key to autocomplete keywords.\
         if not matched:
             print_error('< not found library', lib_name)
             return
+        libs = get_libs_as_dict()
         for name in matched:
+            lib = libs[name]
             print_output('< Keywords of library', name)
-            for keyword in get_lib_keywords(name):
+            for keyword in get_lib_keywords(lib):
                 print_output('   {}\t'.format(keyword['name']),
                              keyword['doc'])
 
@@ -554,6 +594,20 @@ use the TAB keyboard key to autocomplete keywords.\
 
     complete_k = complete_keywords
 
+    def do_docs(self, kw_name):
+        """Get keyword documentation for individual keywords
+
+         d(ocs) [<keyword_name>]
+        """
+
+        for lib in get_libs():
+            for keyword in get_lib_keywords(lib, long_format=True):
+                if keyword['name'].lower() == kw_name.lower():
+                    print(keyword['doc'])
+                    return
+        print("could not find documentation for keyword {}".format(kw_name))
+
+    do_d = do_docs
 
 class DebugLibrary(object):
 
@@ -565,7 +619,7 @@ class DebugLibrary(object):
     def debug(self):
         """Open a interactive shell, run any RobotFramework keywords.
 
-        Keywords seperated by two space or one tab, and Ctrl-D to exit.
+        Keywords separated by two space or one tab, and Ctrl-D to exit.
         """
 
         # re-wire stdout so that we can use the cmd module and have readline
@@ -616,30 +670,40 @@ DEBUG FROM CONSOLE
         return s
 
 
-def shell():
-    """A standalone robotframework shell"""
-
-    import tempfile
-    # ceate test suite file for REPL.
-    source = tempfile.NamedTemporaryFile(prefix='robot_debug',
-                                         suffix='.txt', delete=False)
-    source.write(b'''*** Settings ***
+TEST_SUITE = b'''*** Settings ***
 Library  DebugLibrary
 
 ** test case **
 RFDEBUG REPL
     debug
-''')
-    source.flush()
+'''
 
-    args = '-l None -x None -o None -L None -r None ' + source.name
-    from robot import run_cli
-    rc = run_cli(args.split())
 
-    source.close()
-    if os.path.exists(source.name):
-        os.unlink(source.name)
-    sys.exit(rc)
+def shell():
+    """A standalone robotframework shell"""
+
+    with tempfile.NamedTemporaryFile(prefix='robot-debug-',
+                                     suffix='.txt',
+                                     delete=False) as test_file:
+        try:
+            test_file.write(TEST_SUITE)
+            test_file.flush()
+
+            default_no_logs = '-l None -x None -o None -L None -r None'
+            if len(sys.argv) > 1:
+                args = sys.argv[1:] + [test_file.name]
+            else:
+                args = default_no_logs.split() + [test_file.name]
+            rc = run_cli(args)
+
+            sys.exit(rc)
+        finally:
+            test_file.close()
+            # pybot will raise PermissionError on Windows NT or later
+            # if NamedTemporaryFile called with `delete=True`,
+            # deleting test file seperated will be OK.
+            if os.path.exists(test_file.name):
+                os.unlink(test_file.name)
 
 
 if __name__ == '__main__':
